# from enum import Enum, auto
from .application import Application
from .dex import DexFactory, DexPair
from .network import Network
from .token import Token, TokenAmount
from .user import User
from .wallet import Wallet


# class LimitOrderType(Enum):
#     TrailingStopLoss = auto()
#     StopLoss = auto()
#     TakeProfit = auto()


# class Dex(Enum):
#     Camelot = auto()
#     Uniswap = auto()
#     UniswapV3 = auto()
#     SushiSwap = auto()


__all__ = [
    "Application",
    # "Dex",
    "DexFactory",
    "DexPair",
    # "LimitOrderType",
    "Network",
    "Token",
    "TokenAmount",
    "User",
    "Wallet",
]



import enum
from typing import Optional, Union
from uuid import UUID

from pydantic import BaseModel, Field

from ..utils.client import _force_get_global_client


class WalletType(enum.Enum):
    mnemonic = "mnemonic"  # generated via mnemonic index
    private_key = "private_key"  # custodial (KMS)
    enclave = "enclave"  # enclave
    noncustodial = "noncustodial"  # noncustodial


class WalletAppData(BaseModel):
    archived: bool = False
    notes: dict[str, Union[str, int]] = {}

    def __repr__(self):
        return f"<AppData| archived: {self.archived}, notes: {self.notes}>"

    __str__ = __repr__


class Wallet(BaseModel):
    """
    A wallet, which is able to make transactions.

    Wallets are owned by a specific user, and have individual permissions.
    """

    id: UUID
    name: str
    address: str
    type: WalletType
    group_id: Optional[str] = Field(alias="groupId")
    owner_id: Optional[UUID] = Field(alias="ownerId")
    creator_app_id: Optional[UUID] = Field(alias="creatorAppId")
    data: WalletAppData = WalletAppData()

    def __repr__(self):
        return f"<{self.type.value.capitalize()} Wallet: '0x..{self.address[-4:]}'>"

    @classmethod
    async def load(cls, address):
        """
        If you load a wallet by address, it will give a noncustodial wallet.
        This represents a wallet but has none of the capabilities of a wallet
        with a private key attached to it.
        """
        client = _force_get_global_client()
        wallet = await client.wallet.load(address)
        return cls(**wallet.json())

    async def get_data(self):
        """
        Get app specific data associated with wallet.
        """
        client = _force_get_global_client()
        response = await client.wallet.get_wallet_data(self.id)
        return WalletAppData(**response.json())

    async def update_data(
        self, archive: bool = False, notes: dict[str, Union[int, str]] = {}
    ):
        client = _force_get_global_client()
        response = await client.wallet.update_wallet_data(self.id, archive, notes)
        return WalletAppData(**response.json())



from enum import IntEnum, auto
from uuid import UUID

from pydantic import BaseModel, Field

from .token import Token


class VaultType(IntEnum):
    bank = auto()
    vault = auto()


class Vault(BaseModel):
    app_id: UUID = Field(alias="appId")
    wallet_id: UUID = Field(alias="walletId")
    token: Token
    name: str
    description: str
    type: str
    balance: float
    shares: float



from httpx import Response


class RateLimitError(Exception):
    """Raised when hitting rate limit"""


class NotFoundError(Exception):
    """Information was not found during execution"""


class UnknownError(Exception):
    """Unknown Error.  Please contact Empyreal Team"""


def handle_response_error(response: Response):
    if response.status_code == 429:
        raise RateLimitError(response.json()["detail"])
    elif response.status_code == 400:
        raise NotFoundError(response.json()["detail"])
    elif response.status_code == 500:
        raise UnknownError()


__all__ = [
    "RateLimitError",
    "UnknownError",
]



from contextvars import ContextVar
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from empyrealSDK.sdk import EmpyrealSDK

_current_client: ContextVar[Optional["EmpyrealSDK"]] = ContextVar(
    "_current_client", default=None
)


def _get_global_client() -> Optional["EmpyrealSDK"]:
    return _current_client.get()


def _force_get_global_client():
    if not (client := _get_global_client()):
        raise ValueError("Must initialize EmpyrealSDK client first")
    return client


def _set_global_client(c: Optional["EmpyrealSDK"]) -> None:
    if c is not None:
        _current_client.set(c)


def _del_global_client(c: "EmpyrealSDK") -> None:
    if c == _current_client.get():
        _current_client.set(None)



from .application import ApplicationResource
from .infra import PingResource
from .token import TokenResource
from .user import UserResource
from .wallet import WalletResource
from .vault import VaultResource

__all__ = [
    "ApplicationResource",
    "PingResource",
    "TokenResource",
    "UserResource",
    "VaultResource",
    "WalletResource",
]



from typing import Optional
from uuid import UUID

from empyrealSDK.utils import RequestHelpers
from empyrealSDK.types import Application
from empyrealSDK.exc import handle_response_error


class ApplicationResource(RequestHelpers):
    async def info(self):
        """Get basic info about the app given the provided SDK api_key

        Returns:
            Application: a dataclass containing the general information
        Raises:
            ValueError: If api key is invalid
        """
        response = await self._get("app/")
        if not response.status_code == 200:
            raise ValueError(response.json()["detail"])
        return Application(**response.json())

    async def update(
        self,
        swap_fee: Optional[int] = None,
        fee_collection_amount: Optional[int] = None,
        app_wallet_id: Optional[UUID] = None,
    ):
        """update app config.  Any fields left as `None` will be ignored.

        :returns: bool: If the update was successful or not
        """

        response = await self._put(
            "app/",
            json={
                "swapFee": swap_fee,
                "feeCollectionAmount": fee_collection_amount,
                "appWalletId": app_wallet_id,
            },
        )
        return response

    async def update_api_key(
        self,
    ):
        response = await self._put(
            "app/apikey",
        )
        handle_response_error(response)
        new_api_key = response.json()["apiKey"]
        self.sdk.api_key = new_api_key

        return new_api_key



from typing import Optional
from uuid import UUID

from eth_typing import ChecksumAddress, HexStr
from httpx import Response

from empyrealSDK.utils import RequestHelpers
from empyrealSDK.exc import handle_response_error


class WalletResource(RequestHelpers):
    async def load(self, address: ChecksumAddress):
        response = await self._post(
            "wallets/noncustodial",
            json={
                "address": address,
            },
        )
        return response

    async def get_app_wallets(self, user_id: UUID) -> Response:
        response = await self._get("wallets/app", params={"userId": str(user_id)})
        handle_response_error(response)
        return response

    async def archive(self, wallet_id: UUID):
        response = await self._put(
            "wallets/archive",
            json={
                "walletId": wallet_id,
            },
        )
        handle_response_error(response)
        return response

    async def make_wallet(
        self,
        user_id: UUID,
        name: str,
        private_key: Optional[HexStr] = None,
    ):
        if private_key:
            return await self._make_pk_wallet(user_id, name, private_key)
        return await self._make_mnemonic_wallet(user_id, name)

    async def update_wallet_data(
        self,
        wallet_id: UUID,
        archive: bool = False,
        notes: dict[str, str] = {},
    ):
        return await self._put(
            "wallets/data",
            json={
                "walletId": str(wallet_id),
                "archived": archive,
                "notes": notes,
            },
        )

    async def get_wallet_data(
        self,
        wallet_id: UUID,
    ):
        return await self._get(
            f"wallets/data/{wallet_id}",
        )

    async def _make_mnemonic_wallet(self, user_id: UUID, name: str) -> Response:
        response = await self._post(
            "wallets/mnemonic",
            json={
                "userId": str(user_id),
                "name": name,
            },
        )
        handle_response_error(response)
        return response

    async def _make_pk_wallet(
        self,
        user_id: UUID,
        name: str,
        private_key: HexStr,
    ) -> Response:
        response = await self._post(
            "wallets/pk",
            json={
                "userId": str(user_id),
                "name": name,
                "privateKey": private_key,
            },
        )
        handle_response_error(response)
        return response



from functools import singledispatchmethod
from typing import Literal, Optional, Union
from uuid import UUID

from eth_typing import ChecksumAddress, HexStr
from pydantic import BaseModel, Field

from .network import Network
from .wallet import Wallet
from ..utils.client import _force_get_global_client


class Token(BaseModel):
    """An abstraction of an ERC20 token instance"""

    id: UUID = Field()
    address: ChecksumAddress
    name: str
    symbol: str
    decimals: int
    network: Network = Field(alias="chainId")

    @classmethod
    async def load(
        cls,
        address: ChecksumAddress,
        network: Network = Network.Ethereum,
    ):
        client = _force_get_global_client()
        response = await client.token.lookup(address, network.value)
        return cls(**response.json())

    async def allowance(
        self,
        owner: ChecksumAddress,
        spender: ChecksumAddress,
        block_num: Optional[Union[int, Literal["latest"]]] = "latest",
    ) -> "TokenAmount":
        """
        Gets the allowance allocated to an address from a spender


        :param owner: A checksummed ethereum address
        :param spender: A checksummed ethereum address
        :return: :class:`.TokenAmount`
        """
        client = _force_get_global_client()
        allowance = await client.token.allowance(
            self.address,
            owner,
            spender,
            self.network.value,
            block_num=block_num,
        )
        return TokenAmount(
            amount=allowance,
            decimals=self.decimals,
            token=self,
        )

    async def approve(
        self,
        from_wallet: Wallet,
        spender: Union[Wallet, ChecksumAddress],
        amount: int = int(2**256 - 1),
        priority_fee: Optional[int] = None,
    ) -> HexStr:
        """
        Approve a spender to use a token.
        :param from_wallet: :class:`empyrealSDK.Wallet` making the approval
        :param spender: A checksummed ethereum address
        :return: HexStr

        """
        client = _force_get_global_client()
        if isinstance(spender, Wallet):
            return await client.token.approve(
                self.id,
                from_wallet.id,
                spender.address,
                chain_id=self.network.chain_id,
                amount=amount,
                priority_fee=priority_fee,
            )
        return await client.token.approve(
            self.id,
            from_wallet.id,
            spender,
            chain_id=self.network.chain_id,
            amount=amount,
            priority_fee=priority_fee,
        )

    async def transfer(
        self,
        from_wallet: Wallet,
        recipient: Union[Wallet, ChecksumAddress],
        amount: "TokenAmount",
        gas_price: Optional[int] = None,
    ) -> HexStr:
        """
        Transfer a token amount to a target address

        :param wallet: :class:`empyrealSDK.Wallet`
        :param recipient: The recipient of the transfer
        :param amount: Amount of tokens to transfer
        :return: HexStr of the transaction
        """
        client = _force_get_global_client()
        if isinstance(recipient, Wallet):
            recipient_address = recipient.address
        else:
            recipient_address = recipient
        return await client.token.transfer(
            self.id,
            from_wallet.id,
            recipient_address,
            amount,
            gas_price=gas_price,
        )

    @singledispatchmethod
    async def balance_of(
        self,
        wallet: Wallet,
        block: Union[int, Literal["latest"]] = "latest",
    ) -> "TokenAmount":
        """
        Gets the balance of a wallet at a particular block.
        Defaults to latest block.

        :param wallet: :class:`empyrealSDK.Wallet`
        :param block: A block number of defaults to latest block
        :return: :class:`.TokenAmount` of the current balance
        """

        client = _force_get_global_client()
        balance = await client.token.balance_of(
            self.address,
            wallet.address,
            self.network.value,
            block,
        )
        return TokenAmount(amount=balance, decimals=self.decimals, token=self)

    @balance_of.register(str)
    async def _(
        self,
        wallet_address: ChecksumAddress,
        network: Network = Network.Ethereum,
        block: Union[int, Literal["latest"]] = "latest",
    ):
        """
        Gets the balance of a wallet at a particular block,
        given a Checksum Address.

        :param wallet: :class:`empyrealSDK.Wallet`
        :param block: A block number of defaults to latest block
        :return: :class:`.TokenAmount` of the current balance
        """

        client = _force_get_global_client()
        balance = await client.token.balance_of(
            self.address,
            wallet_address,
            network.value,
            block,
        )
        return TokenAmount(
            amount=balance,
            decimals=self.decimals,
            token=self,
        )

    def __repr__(self):
        return f"<'{self.symbol}' on {self.network.name}>"

    __str__ = __repr__


class TokenAmount(BaseModel):
    """An abstraction on a token amount"""

    amount: int
    decimals: int = 18
    token: Optional["Token"] = None
    price: Optional[int] = None

    def format(self, num_decimals=2):
        return round(self.amount / 10**self.decimals, num_decimals)

    def __truediv__(self, other):
        return self._div_by_other(other)

    def __mul__(self, other):
        return self._mul_other(other)

    @singledispatchmethod
    def _mul_other(self, other: int):
        return TokenAmount(
            amount=self.amount // other,
            decimals=self.decimals,
            token=self.token,
            price=self.price,
        )

    @singledispatchmethod
    def _div_by_other(self, other: int):
        return TokenAmount(
            amount=self.amount // other,
            decimals=self.decimals,
            token=self.token,
            price=self.price,
        )

    def __repr__(self):
        if self.token:
            return f"<{self.token.name}: {self.amount / 10**self.decimals}>"
        return f"<Amount: {self.amount / 10**self.decimals}>"

    __str__ = __repr__


@TokenAmount._div_by_other.register(TokenAmount)  # type: ignore
def _(self: TokenAmount, other: TokenAmount):
    return TokenAmount(
        amount=int(self.amount / (other.amount / 10**other.decimals)),
        decimals=self.decimals,
        token=self.token,
        price=self.price,
    )


@TokenAmount._mul_other.register(TokenAmount)  # type: ignore
def _(self: TokenAmount, other: TokenAmount):
    return TokenAmount(
        amount=int(int(self.amount * other.amount) / 10**other.decimals),
        decimals=self.decimals,
        token=self.token,
        price=self.price,
    )



#!/usr/bin/env python
from setuptools import (
    setup,
)

with open("README.md") as f:
    readme = f.read()

extras_require = {
    "linter": [
        "black>=22.1.0",
        "flake8==3.8.3",
        "isort>=5.11.0",
        "mypy>=1.0.0",
        "types-setuptools>=57.4.4",
    ],
    "dev": [
        "pytest>=7.0.0",
        "pytest-asyncio>=0.18.1",
        "pytest-mock>=1.10",
        "pytest-watch>=4.2",
        "pytest-xdist>=1.29",
        "setuptools>=38.6.0",
    ],
}

setup(
    name="empyrealSDK",
    version="0.0.1.post7",
    description="Empyreal SDK for web3 development",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Empyreal",
    author_email="dev@empyreal.app",
    url="https://github.com/empyrealapp/empyrealSDK",
    install_requires=[
        "eth-typing>=3.0.0",
        "eth-utils>=2.1.0",
        "hexbytes>=0.1.0",
        "httpx>=0.23.3",
        "pydantic>=2.3.0",
    ],
    python_requires=">=3.7.2",
    extras_require=extras_require,
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
)



from typing import Literal

from .modules.dex import price

from .modules import core
from .utils.client import _set_global_client

ENVIRONMENTS = {
    "local": "http://localhost:8080",
    "prod": "https://api.empyrealsdk.com",
}


class EmpyrealSDK:
    """
    The EmpyrealSDK instance.  This is currently a global, singleton instance.
    So when a user initializes the empyrealSDK instance, this will be used automatically
    in all Empyreal types.

    Examples
    --------
    Initially in any codebase using the SDK, be sure to declare

    >>> from empyrealSDK import *
    >>> api_key = "<api_key>"
    >>> EmpyrealSDK(api_key)
    >>> app = Application.load()

    This will declare your SDK instance and inject it in all method invocations.
    Then when you call 5 :meth:`empyrealSDK.types.Application.load`, you are able to get the instance attached
    to your ``API_KEY``.
    """

    rpc_url: str
    api_key: str

    def __init__(self, api_key: str, env: Literal["local", "prod"] = "prod"):
        if env not in ENVIRONMENTS:
            raise ValueError(
                "Invalid Environment.  Must provide one of ['local', 'prod']"
            )
        self.rpc_url = ENVIRONMENTS[env]
        self.api_key = api_key

        self.app = core.ApplicationResource(self)
        self.infra = core.PingResource(self)
        self.token = core.TokenResource(self)
        self.user = core.UserResource(self)
        self.vault = core.VaultResource(self)
        self.wallet = core.WalletResource(self)
        self.prices = price.PriceResource(self)
        _set_global_client(self)



from typing import Any, Mapping, Optional, TYPE_CHECKING

import httpx
from httpx import Response
from httpx._types import PrimitiveData

if TYPE_CHECKING:
    from .. import EmpyrealSDK


class RequestHelpers:
    def __init__(self, sdk: "EmpyrealSDK", version="v1"):
        self.sdk = sdk
        self.version = version

    @property
    def rpc_url(self):
        return self.sdk.rpc_url

    @property
    def api_key(self):
        return self.sdk.api_key

    async def _get(
        self, path: str, params: Optional[Mapping[str, PrimitiveData]] = None
    ) -> Response:
        async with httpx.AsyncClient() as client:
            response = await client.get(
                f"{self.rpc_url}/{self.version}/{path}",
                headers={
                    "API-KEY": self.api_key,
                },
                params=params,
            )
        return response

    async def _post(self, path: str, json: Any) -> Response:
        async with httpx.AsyncClient() as client:
            return await client.post(
                f"{self.rpc_url}/{self.version}/{path}",
                json=json,
                headers={
                    "API-KEY": self.api_key,
                },
            )

    async def _put(self, path: str, json: Any = {}) -> Response:
        async with httpx.AsyncClient() as client:
            return await client.put(
                f"{self.rpc_url}/{self.version}/{path}",
                json=json,
                headers={
                    "API-KEY": self.api_key,
                },
            )

    async def _delete(self, path: str) -> Response:
        async with httpx.AsyncClient() as client:
            return await client.delete(
                f"{self.rpc_url}/{self.version}/{path}",
                headers={
                    "API-KEY": self.api_key,
                },
            )
